# Functions used by makedist scripts
#
# Copyright (c) 2004-2020 by Silvan Calarco (silvan.calarco@mambasoft.it)
# Copyright (c) 2004-2005 by Davide Madrisan (davide.madrisan@gmail.com)
#

# final optimizations: prelink and pre-create ldconfig cache
function prelink() {
   echo "Performing binaries prelinking..."
   chroot $MOUNTDIR /bin/bash -c "PRELINK_NOINITCHECK=1 /etc/cron.daily/prelink"
   echo "Pre-creating ldconfig cache..."
   chroot $MOUNTDIR /sbin/ldconfig
}

function i18n_add_all_languages() {
   # add localedata for all supported languages (requires glibc-i18ndata installed)
   echo -n "Generating locales archive for all supported languages..."
   cat $MOUNTDIR/usr/share/i18n/locales/SUPPORTED | sed "s|\\\||" | \
   while read line; do
      LOCALE=${line/\/*}
      LOCALEPREFIX=${LOCALE/.*}
      ENCODING=${line/*\/}
      [ "$LOCALE" != "$ENCODING" ] || continue
      chroot $MOUNTDIR localedef -i $LOCALEPREFIX -f $ENCODING $LOCALE
      echo -n "$line "
   done
   echo
}

function i18n_add_languages_from_file() {
   # add localedata for extra requested languages from input file (requires glibc-i18ndata installed)
   echo -n "Generating locales archive for requested languages..."
   cat $TARGETDIR/platforms/$PLATFORM/languages | \
   while read REQUESTED_LANG; do
      lines=`grep -i "^${REQUESTED_LANG}[./@]" $MOUNTDIR/usr/share/i18n/locales/SUPPORTED`
      if [ ! "$lines" ]; then
         echo
         echo "Warning: requested language $REQUESTED_LANG not found in supported languages; skipping."
      else
         echo $lines | sed "s|\\\||g" | \
         for line in $lines; do
            LOCALE=${line/\/*}
            LOCALEPREFIX=${LOCALE/.*}
            ENCODING=${line/*\/}
            [ "$LOCALE" != "$ENCODING" ] || continue
            chroot $MOUNTDIR localedef -i $LOCALEPREFIX -f $ENCODING $LOCALE
            echo -n "$line "
         done
      fi
   done
   echo
}

function add_additional_defs_file() {
   local ADDITIONAL_DEFS_FILE=$1

   [ -e ${MOUNTDIR}/${ADDITIONAL_DEFS_FILE} ] || return 1

   # switching to using internal 'rpm -q' is problematic when cross building, so clean it
   clean_rpm_database

   echo "Adding additional defs file $1 from $MOUNTDIR to Makefile..."

   local tmpfile=`mktemp -q -t makedist.XXXXXXXX.tmp`
   chroot $MOUNTDIR sh -c "set > /tmp/set1; . $ADDITIONAL_DEFS_FILE; set > /tmp/set2"
   awk 'FNR==NR{old[$0];next};!($0 in old)' ${MOUNTDIR}/tmp/set1 ${MOUNTDIR}/tmp/set2 > $tmpfile
   rm -f $MOUNTDIR/tmp/set1 $MOUNTDIR/tmp/set2

   . ${tmpfile}
   cat ${tmpfile} | \
   while read line; do 
      if [[ "$line" =~ ([A-Z0-9_]*)= ]]; then
         echo "${BASH_REMATCH[1]}: ${!BASH_REMATCH[1]}" >> $LOCALSTATEDIR/.Makefile.inc
      fi
   done
   rm -f ${tmpfile}
}

function create_devices() {
echo "Creating devices in /dev..."
#mkdir -p $MOUNTDIR/dev
(cd $MOUNTDIR/dev
  ln -sf /proc/self/fd fd
  ln -sf /proc/self/fd/0 stdin
  ln -sf /proc/self/fd/1 stdout
  ln -sf /proc/self/fd/2 stderr
  ln -sf /proc/kcore core
  mkdir shm pts
  mknod --mode=0600 initctl p
  mknod -m666 null c 1 3
  mknod -m600 console c 5 1
  mknod -m444 urandom c 1 9
)
}

function initialize_root_dir() {
   echo "Mounting tmpfs on MOUNTDIR (size=$MOUNTDIR_TMPFS_SIZE)..."
   [ "$MOUNTDIR" == "" -o "$MOUNTDIR" == "/" ] &&  {
      echo "$makedist_me: dangerous: MOUNTDIR is empty, aborting."
      exit 1;
   }

   [ "$MOUNTDIR_TMPFS" ] && umount $MOUNTDIR_TMPFS
   MOUNTDIR_TMPFS=$MOUNTDIR
   umount $MOUNTDIR 2>/dev/null
   mount tmpfs $MOUNTDIR -t tmpfs -o size=$MOUNTDIR_TMPFS_SIZE
   if [ $? -ne 0 ]; then
      echo "Warning: error $? mounting tmpfs on $MOUNTDIR"
      MOUNTDIR_TMPFS=
   fi

   mkdir $MOUNTDIR/proc
   mount -o bind /proc $MOUNTDIR/proc

   mkdir $MOUNTDIR/dev
   mount none -t tmpfs $MOUNTDIR/dev

   mkdir $MOUNTDIR/run
   mount -o bind /run $MOUNTDIR/run

   mkdir $MOUNTDIR/sys
   mount -o bind /sys $MOUNTDIR/sys

   create_devices

   # / to /usr symlinks
   mkdir -p $MOUNTDIR/usr/{bin,sbin,lib}
   ln -s usr/bin $MOUNTDIR/bin
   ln -s usr/sbin $MOUNTDIR/sbin
   ln -s usr/lib $MOUNTDIR/lib

   if [ "${arch}" = "x86_64" -o "${arch}" = "aarch64" ]; then
      mkdir -p $MOUNTDIR/usr/lib64
      ln -s usr/lib64 $MOUNTDIR/lib64
   fi

   #[ $MEDIA = "raw" ] && create_raw_filesystem $LOCALSTATEDIR/$MEDIA_NAME $MEDIA_SIZE 1024k $ROOTFS
   [ $MEDIA = "raw" ] && create_raw_disk_image $MEDIA_SIZE 1024k

   trap "[ $LOOPDEV ] && losetup -d $LOOPDEV; chroot $MOUNTDIR /usr/bin/gpgconf --kill gpg-agent 2>/dev/null; [ -e $MOUNTDIR/proc/version ] && umount $MOUNTDIR/proc; [ -e $MOUNTDIR/dev/null ] && umount $MOUNTDIR/dev; [ -e $MOUNTDIR/run/systemd ] && umount $MOUNTDIR/run; [ -e $MOUNTDIR/sys/kernel ] && umount $MOUNTDIR/sys; [ \"$MOUNTDIR_TMPFS\" ] && umount $MOUNTDIR_TMPFS" 0 HUP INT QUIT ABRT KILL TERM
}

function create_rpm_database() {
# RPM database initialization
#
   echo "RPM database initialization..."

   [ -e $MOUNTDIR/var/lib/rpm ] && rm -rf $MOUNTDIR/var/lib/rpm

   [ "$MEDIUM_NAME" ] || MEDIUM_NAME=`medium_name`
   BUILDSCRIPT_NAME=${MEDIUM_NAME}.build.sh
   
   if [ "$RPM_CHROOT" ]; then
      RPM_CHROOT_PREFIX="chroot $MOUNTDIR"
      RPM_ROOT_APPEND=
   else
      RPM_CHROOT_PREFIX=
      RPM_ROOT_APPEND="-r $MOUNTDIR"
   fi

   echo "#!/bin/sh
# Build script generated by makedist for ${MAKEDIST_TARGET} ${PLATFORM} platform
#
# Output medium: ${MEDIUM_NAME}

# create_rpm_database():
#
MOUNTDIR=$MOUNTDIR
mkdir -p \$MOUNTDIR/var/lib/rpm
$RPM_CHROOT_PREFIX $RPM $RPM_ROOT_APPEND --initdb" > ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}

   mkdir -p $MOUNTDIR/var/lib/rpm
   $RPM_CHROOT_PREFIX $RPM $RPM_ROOT_APPEND --initdb ||
      { echo "$makedist_me: unable to initialize the RPM database" >&2
         { (exit 1); exit 1; }; }
   [ -e "$RPM_GPG_KEY" ] && {
      $RPM_CHROOT_PREFIX $RPM $RPM_ROOT_APPEND --import $RPM_GPG_KEY
      echo "$RPM_CHROOT_PREFIX $RPM $RPM_ROOT_APPEND --import $RPM_GPG_KEY" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}
   }
}

function clean_rpm_database() {
   echo "Cleaning rpm database"
   [ "$MOUNTDIR" != "" -a "$MOUNTDIR" != "/" ] && rm -rf $MOUNTDIR/var/lib/rpm/__db.*
   # Kill any gpg-agent daemon spawned daemon preventing umount
   chroot $MOUNTDIR /usr/bin/gpgconf --kill gpg-agent || :
}

function erase_rpm_database() {
   echo "Erasing rpm database"
   [ "$RPM_OCCUPACY_STATS_FILE" ] || rpm_occupacy_stats
   [ "$MOUNTDIR" != "" -a "$MOUNTDIR" != "/" ] && rm -rf $MOUNTDIR/var/lib/rpm
}

function rpm_occupacy_stats() {
  RPM_OCCUPACY_STATS_FILE=`mktemp`
  SIZE=0
  rpm -r $MOUNTDIR -qa --queryformat "%{SIZE} %{NAME}\n" | sort -k1 -n -r > $RPM_OCCUPACY_STATS_FILE
  while read line; do
     set -- $line
     SIZE=`expr $SIZE + $1`
  done < $RPM_OCCUPACY_STATS_FILE
#  echo "-------------------------" >> $RPM_OCCUPACY_STATS_FILE
#  echo "$SIZE \tTotal size" >> $RPM_OCCUPACY_STATS_FILE
}

function set_rpm_dirs() {

   for r in ${RPMDIR[*]}; do
      RPM_DIRS="${r}${arch}/ $RPM_DIRS"
   done

   if [ "$arch_fallback" ]; then
      for r in ${RPMDIR[*]}; do
         RPM_DIRS="$RPM_DIRS ${r}${arch_fallback}/"
      done
   fi
}

# prints a list of rpms full pathnames given the simple names
function expand_rpm_names() {
   local i j a

   set_rpm_dirs

   for i in $1; do
      a=`find $RPM_DIRS -maxdepth 1 -regex ".*/${i//+/\\\+}-[^-]*-[^-]*\.rpm" | \
        grep -m 1 ""`
      [ "$a" ] && echo -n "$a " || {
         if [ "$BUILD_MISSING" -a "$BUILD_MISSING" != "0" ]; then
            resolve_pkg_builddeps $i
            for j in $PKGS_RESOLVED; do
               a=`find $RPM_DIRS -maxdepth 1 -regex ".*/${j//+/\\\+}-[^-]*-[^-]*\.rpm" | \
                 grep -m 1 ""`
               [ "$a" ] || {
                  echo "Porting package $j required by $i" > /dev/stderr
                  su -l $PORT_AUTODIST_USER -c "autodist -p -b $j -s --server $PORT_DEST_REPOSITORY --arch ${arch} --force" 1>&2 || {
                     echo "Error porting $i package; aborting." > /dev/stderr
                     return 1;
                  }
               }
            done
         else
            echo
            echo "Error expanding package '$i': expanded rpm not found; aborting." > /dev/stderr
	    echo "Dirs:$RPM_DIRS" > /dev/stderr
	    return 1;
	 fi
      }
   done
}

function expand_srpm_names() {
   local i a

   for r in ${SRPMDIR[*]}; do
      SRPM_DIRS="$r $SRPM_DIRS"
   done

   for i in $1; do
      a=`find $SRPM_DIRS -maxdepth 1 -regex ".*/${i//+/\\+}-[^-]*-[^-]*" | \
        grep -m 1 ""`

      [ "$a" ] && echo $a || {
            echo
            echo "Error expanding package $i: expanded srpm not found; aborting." > /dev/stderr
	    echo "Dirs:$SRPM_DIRS" > /dev/stderr
	    return 1;
      }
   done
}

# prints the version of the rpm given
function get_rpm_version() {

   set_rpm_dirs
   find $RPM_DIRS -maxdepth 1 -regex ".*/${1}-[^-]*-[^-]*" | \
      grep -m 1 "" | \
      sed "s|.*/${1}-\([^-]*-[0-9]*\)[^-]*|\1|"
   #find $RPMDIR -regex ".*/${1//+/\\+}-[^-]*-[^-]*" | sed "s|.*/${1//+/\\+}-\([^-]*\)-[^-]*|\1|"
   #find $NOARCHDIR -regex ".*/${1//+/\\+}-[^-]*-[^-]*" | sed "s|.*/${1//+/\\+}-\([^-]*\)-[^-]*|\1|"
}


# wget_grep()
#
# parameters:
#    $1 : lenght of the first column in chars
#    $2 : \hfill character (e.g.: '.'); default is ' '
# description:
#    This function greps the output of a wget command and returns
#    more compact strings that looks like:
#      bash-2.05b-4qilnx.i586.rpm.................. bytes: 48.543
#                                <-- fill-chars -->
#      <--------------- $1 chars ----------------->
function wget_grep() {
   sed -n "/.* URL: .*/!d;/-> .*\.listing.*\[.*\]$/d  # discard unneeded lines
      s/.*\/\(.*\)\[\([0-9]*\)].\+$/\1 \2/  # select file name and size
      # modify file size to a dot separated string: e.g. 12345 --> 12.345
      # and copy all (name + modified-size) to the hold space
      :a s/\(.*\)\(.*[0-9]\)\([0-9]\{3\}\)/\1\2.\3/g;ta;h
      s/\([^ ]*\).*/\1 /  # now select from it the file name...
      # ... put it in a box $dtab long, and append all to hold space
      :b s/^.\{1,$1\}$/&${2:- }/;tb;G
      # remove '\n'filename, put 'bytes:' between filled-filename and size,
      # print the resulting pattern space
      s/^\(.*\)\n[^ ]* \([0-9.]*\)/\1 bytes:\2/p"
}


# install_kernel()
#
# parameters:
#    $1 : kernel extraver
function install_kernel() {
   # installing kernel image, modules and utilities
   KERNEL_BASENAME="kernel"

   [ "$1" ] && {
      KERNEL_TARGET="$KERNEL_BASENAME-$1"
      for a in $2; do
         KERNEL_TARGET="$KERNEL_TARGET kernel-$1-$a"
      done
   }

   if [ "$KERNEL_NOPREPARATION" != "1" ]; then
      install_rpms_by_target "\
Installing and preparing dracut for kernel install..." \
      "dracut module-init-tools"

      for f in /usr/sbin/dracut /usr/sbin/mkinitrd /sbin/depmod; do
         mv ${MOUNTDIR}${f} ${MOUNTDIR}${f}.preinst
         ln -s /bin/true ${MOUNTDIR}${f}
      done
   fi

   MAKE_RESOLVE_ADD="KERNEL_EXTRAVER=$1" install_rpms_by_target "\
Installing kernel image (${KERNEL_VER}$1), modules and utilities..." \
      "$KERNEL_TARGET"
   MAKE_RESOLVE_ADD=

   if [ "$KERNEL_NOPREPARATION" != "1" ]; then
      for f in /usr/sbin/dracut /usr/sbin/mkinitrd /sbin/depmod; do
         rm ${MOUNTDIR}${f}
         mv ${MOUNTDIR}${f}.preinst ${MOUNTDIR}${f}
      done
   fi

}

# update_modules_dependencies()
#
# parameters:
#    $1 : $KERNEL_VER$KERNEL_EXTRAVER (e.g. 2.4.27preempt)
function update_modules_dependencies() {
   [ -e ${MOUNTDIR}/lib/modules/$1 ] && {
      $CHROOTCMD depmod -a -F /boot/System.map-$1 $1
      touch $MOUNTDIR/lib/modules/$1/kernel/modules.dep
   }
}

# calculate_rpms_size
#
# parameters:
#
# $1 : list of RPMS
function calculate_rpms_occupation() {

   rpmsfilesum=0
   RPMS_FILELIST=`expand_rpm_names "$1"` || exit 1
   for rpm in $RPMS_FILELIST; do
      filesize=`du -b $rpm | gawk '{ print $1 }'`
      rpmsfilesum=`expr $rpmsfilesum + $filesize`
   done
}

function pre_qemu() {
   case $ARCH in
      i386|i486|i586|i686|athlon) QEMUARCH=i386 ;;
      *) QEMUARCH=$ARCH
   esac
   echo "Configuring qemu-$QEMUARCH user mode emulator for RPM installation scripts"
   [ -e /usr/bin/qemu-$QEMUARCH.static ] || {
      echo "Error: missing qemu static user mode emulator for $QEMUARCH architecture; aborting."
      exit 1
   }
   # FIXME: may cause conflicts when qemu is installed to target but as now it is not the case
   echo "#
# pre_qemu()
#
# install qemu
cp /usr/bin/qemu-$QEMUARCH.static \$MOUNTDIR/usr/bin/qemu-$QEMUARCH
" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}
   cp /usr/bin/qemu-$QEMUARCH.static $MOUNTDIR/usr/bin/qemu-$QEMUARCH
}

function post_qemu() {
   [ -e $MOUNTDIR/usr/bin/qemu-$QEMUARCH ] && {
      echo "#
# post_qemu()
rm -f \$MOUNTDIR/usr/bin/qemu-$QEMUARCH
" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}
      rm -f $MOUNTDIR/usr/bin/qemu-$QEMUARCH
   }
}

function pre_install_rpms() {

   # disable the service command
   mv $MOUNTDIR/sbin/service $MOUNTDIR/sbin/service.preinst
   ln -sf /bin/true $MOUNTDIR/sbin/service

   echo "#
# pre_install_rpms():
#
# disable the service command
mv \$MOUNTDIR/sbin/service \$MOUNTDIR/sbin/service.preinst
ln -sf /bin/true \$MOUNTDIR/sbin/service
" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}

   # disable the install-info command if --excludedocs is set
   if [ "${rpmextraargs/--excludedocs}" != "${rpmextraargs}" ]; then
      mv $MOUNTDIR/sbin/install-info $MOUNTDIR/sbin/install-info.preinst
      ln -sf /bin/true $MOUNTDIR/sbin/install-info
      echo "#
# disable install-info
mv \$MOUNTDIR/sbin/install-info \$MOUNTDIR/sbin/install-info.preinst
ln -sf /bin/true \$MOUNTDIR/sbin/install-info
" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}
   fi
}

function post_install_rpms() {

   # restore the service command
   #
   rm -f $MOUNTDIR/sbin/service
   mv $MOUNTDIR/sbin/service.preinst $MOUNTDIR/sbin/service

   echo "#
# post_install_rpms():
#
# restore the service command
rm -f \$MOUNTDIR/sbin/service
mv \$MOUNTDIR/sbin/service.preinst \$MOUNTDIR/sbin/service
" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}

   if [ "${rpmextraargs/--excludedocs}" != "${rpmextraargs}" ]; then
      rm -f $MOUNTDIR/sbin/install-info
      mv $MOUNTDIR/sbin/install-info.preinst $MOUNTDIR/sbin/install-info
   echo "#
# restore the install-info command
rm -f \$MOUNTDIR/sbin/install-info
mv \$MOUNTDIR/sbin/install-info.preinst \$MOUNTDIR/sbin/install-info
" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}
   fi

   [ "$RPM_OCCUPACY_STATS_FILE" ] || rpm_occupacy_stats

}

# install_rpms()
#
# parameters:
#    $1 : list of rpms to install
#    $2 : string passed as RPM options (optional)

function install_rpms() {

   TOINSTALL_RPMS=`expand_rpm_names "$1"` || exit 1
   TOINSTALL_SIZE=

   if [ "$RPM_CHROOT" ]; then
      RPM_CHROOT_PREFIX="chroot $MOUNTDIR"
      RPM_ROOT_APPEND=
      cp $TOINSTALL_RPMS $MOUNTDIR/tmp/
      TOINSTALL_RPMS_NEW=
      for p in $TOINSTALL_RPMS; do
         TOINSTALL_RPMS_NEW="$TOINSTALL_RPMS_NEW /tmp/`basename $p`"
      done
      TOINSTALL_RPMS="$TOINSTALL_RPMS_NEW"
   else
      RPM_CHROOT_PREFIX=
      RPM_ROOT_APPEND="-r $MOUNTDIR"
   fi

#   rm $MOUNTDIR/var/lib/rpm/__db.00* -f

   [ "$VERBOSE" ] && {
      echo
      echo "Installing RPMS: $TOINSTALL_RPMS" >&2
   }

   echo "#
# install_rpms():
LANG=${LANGUAGE}_${COUNTRY}.UTF-8 $RPM_CHROOT_PREFIX $RPM --ignorearch $RPM_ROOT_APPEND -hiv $rpmextraargs $2 $TOINSTALL_RPMS" >> ${LOCALSTATEDIR}/${BUILDSCRIPT_NAME}
   LANG=${LANGUAGE}_${COUNTRY}.UTF-8 $RPM_CHROOT_PREFIX $RPM --ignorearch $RPM_ROOT_APPEND -hiv $rpmextraargs $2 $TOINSTALL_RPMS ||
    { retcode=$?
      echo "$makedist_me: fatal error, aborting installation" >&2
      echo "$makedist_me: exit code was: $retcode" >&2
      echo "$makedist_me: command was: LANG=${LANGUAGE}_${COUNTRY}.UTF-8 $RPM_CHROOT_PREFIX $RPM --ignorearch $RPM_ROOT_APPEND -hiv $rpmextraargs $2 $TOINSTALL_RPMS" >&2
      echo "$makedist_me: sleeping 600 seconds"
      sleep 600
      { (exit 1); exit 1; }; }

   INSTALLED="$INSTALLED $1"

   if [ "$RPM_CHROOT" ]; then
      $RPM_CHROOT_PREFIX rm -f $TOINSTALL_RPMS
   fi

}

# resolve_pkg_deps()
#
# the primitive that executes make to resolve dependencies
#
# $1: INSTALLED packages
# $2: TARGETS
#
# Returns:
#
function resolve_pkg_deps() {

   local tmpfile

   tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX` ||
    { echo "${0##*/}: error: cannot create temporary files." >&2
       { (exit 1); exit 1; }; }

   cmd="make --no-print-directory ARCH=$arch -C $GLOBALDIR CHROOTCMD=\"$CHROOTCMD\" \
        LANGUAGE=$LANGUAGE INSTALLED=\"$1\" TARGET=$MAKEFILE_TARGET PLATFORM='$PLATFORM' $MAKE_RESOLVE_ADD -s $2"
   [ "$VERBOSE" ] && {
      echo
      echo "Executing command: $cmd" >&2
   }

   PKGS_RESOLVED="`eval $cmd 2>$tmpfile`"

   if [ $? -ne 0 -a ! "$RESOLVE_PKG_DEPS_NOERROR" ]; then
      MISSING_TARGET=`grep 'No rule to make target' $tmpfile | sed "s|.*target \\\`\(.*\)', needed by.*|\1|"`
      MISSING_TARGET=${MISSING_TARGET/_unresolved_}
      MISSING_TARGET=${MISSING_TARGET/__missing_provider_for_}
      echo "Unresolved dependency for target $2: $MISSING_TARGET"
      expand_rpm_names "$MISSING_TARGET" || exit 1
      echo "FATAL ERROR: error while executing make for target $MAKEFILE_TARGET:"
      cat $tmpfile
      echo "command was: $cmd"
      rm -f $tmpfile
      exit 1
   fi
   rm -f $tmpfile
}

# resolve_pkg_builddeps()
#
# the primitive that executes make to resolve build dependencies
#
# $1: TO_INSTALL packages
#
# Returns:
#
function resolve_pkg_builddeps() {

    local tmpfile

    TOINSTALL=$1

    tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX` ||
    { echo "${0##*/}: error: cannot create temporary files." >&2
       { (exit 1); exit 1; }; }

   cmd="make --no-print-directory PORT_FROM_ARCH=$PORT_FROM_ARCH ARCH=$arch -C $GLOBALDIR SOURCESMODE=1 -s $TOINSTALL"

   [ "$VERBOSE" ] && {
      echo
      echo "Executing command: $cmd" >&2
   }

   PKGS_RESOLVED="`eval $cmd 2>$tmpfile`"

   if [ $? -ne 0 ]; then
      echo "FATAL ERROR: error while executing make for target $MAKEFILE_TARGET:"
      cat $tmpfile
      echo "command was: $cmd"
      rm -f $tmpfile
      exit 1
   fi
   rm -f $tmpfile
}

# install_rpms_by_target()
#
# parameters:
#    $1 : description
#    $2 : makefile target
#    $3 : string passed as RPM options (optional)
function install_rpms_by_target() {

   echo $1

   resolve_pkg_deps "$INSTALLED" "$2"

   TOINSTALL="$PKGS_RESOLVED"

   RESOLVE_PKG_DEPS_NOERROR=1
   if [ "$INSTALL_DEVEL_PACKAGES" ]; then
      for p in $TOINSTALL; do
         resolve_pkg_deps "$INSTALLED $TOINSTALL" "$p-devel"
         TOINSTALL="$TOINSTALL $PKGS_RESOLVED"
      done
   fi
   unset RESOLVE_PKG_DEPS_NOERROR

   [ "$VERBOSE" ] && {
      echo
      #[ $VERBOSE -ge 2 ] && cat $tmpfile
      echo "Installing packages: $TOINSTALL" >&2
   }

   if [ "$TOINSTALL" ]; then
      install_rpms "$TOINSTALL" "$3" 
   else
      echo "Warning: no packages to install for target $2"
   fi
}

function exec_distromatic() {
   distromatic -t distromatic -c ${LOCALSTATEDIR}/distromatic/distromatic.conf --gendatatables -a $arch || {
      echo "Error executing distromatic; command was:"
      echo "distromatic -t distromatic -c ${LOCALSTATEDIR}/distromatic/distromatic.conf --gendatatables -a $arch"
      exit 1
   }
}

# get_pkg_buildinfo - uses distromatic generated build file for
#                     getting information on the repository
# 
# $1: pkg name
function get_pkg_buildinfo() {

   local pkg
   
   [ $1 ] && pkg=$1 || {
      echo "Error: get_pkg_buildinfo: pkg not given; aborting."
      exit 1
   }
   
   [ -e $BUILDSFILE-$arch ] || {
      echo "Error: get_pkg_buildinfo: missing distromatic generated file $BUILDSFILE-$arch; aborting."
      exit 1;
   }

   pkg_header=();
   pkg_builds=();
   pkg_obsoletes=();
   . ${BUILDSFILE}-${arch}.sh

   pkg_name=${pkg_header[0]};
   pkg_arch=${pkg_header[1]};
   pkg_version=${pkg_header[2]};
   pkg_release=${pkg_header[3]};
   pkg_group=${pkg_header[4]};
   pkg_license=${pkg_header[5]};
   pkg_size=${pkg_header[6]};
   pkg_buildtime=${pkg_header[7]};

   return 0;
}

function add_packages_to_repository() {
   local SOURCES INSTALLED i INSTALL_SRPMS INSTALL_RPMS tmpfile

   tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX` ||
    { echo "${0##*/}: error: cannot create temporary files." >&2
       { (exit 1); exit 1; }; }

   [ "$1" ] && DEST_REPOSITORY=$1
   [ "$2" ] && SOURCES=$2
   [ "$3" ] && INSTALLED=$3

   [ "$DEST_REPOSITORY" ] || DEST_REPOSITORY=$MOUNTDIR/repositories/default

   BUILDS=
   for i in $SOURCES; do
      get_pkg_buildinfo $i
      BUILDS="$BUILDS ${pkg_builds[*]}"
   done

   if [ "$3" != "--nodeps" ]; then
      resolve_pkg_deps "$INSTALLED" "$BUILDS"
      TOINSTALL="$PKGS_RESOLVED"

      # find again all sources also of dependent packages
      cmd="make --no-print-directory ARCH=$arch -C $GLOBALDIR SOURCESMODE=1 -s $TOINSTALL"
      [ "$VERBOSE" ] && { 
         echo 
         echo "Executing command: $cmd" >&2
      }

      SOURCES="`eval $cmd 2>$tmpfile`"
   else
      TOINSTALL=$BUILDS
   fi

   INSTALL_RPMS=`expand_rpm_names "$TOINSTALL"` || exit 1
   INSTALL_SRPMS=`expand_srpm_names "$SOURCES"` || exit 1

   [ -e $DEST_REPOSITORY/SRPMS ] || mkdir -p $DEST_REPOSITORY/SRPMS
   [ -e $DEST_REPOSITORY/RPMS/$arch ] || mkdir -p $DEST_REPOSITORY/RPMS/$arch

   for i in $INSTALL_SRPMS; do
      cp $i $DEST_REPOSITORY/SRPMS/ || {
         echo "error: failed to copy $i to $DEST_REPOSITORY/SRPMS/; aborting."
         exit 1
      }
   done

   for i in $INSTALL_RPMS; do
      cp $i $DEST_REPOSITORY/RPMS/$arch/ || {
         echo "error: failed to copy $i to $DEST_REPOSITORY/RPMS/$arch; aborting."
         exit 1
      }
   done
   [ -e "$tmpfile" ] && rm -f $tmpfile
}

function extract_packages() {
   local SOURCES DEST_REPOSITORY i INSTALL_SRPMS

   [ "$1" ] && DEST_REPOSITORY=$1
   [ "$2" ] && PKGS=$2

   [ "$DEST_REPOSITORY" ] || DEST_REPOSITORY=$MOUNTDIR/sources

   INSTALL_RPMS=`expand_rpm_names "$PKGS"` || exit 1

   [ -d "$DEST_REPOSITORY" ] || mkdir -p $DEST_REPOSITORY

   pushd $DEST_REPOSITORY >/dev/null
   for i in $INSTALL_RPMS; do
      $RPM2CPIO $i | cpio -id --no-absolute-filenames || {
         echo "error: failed to extract $i into $DEST_REPOSITORY using $RPM2CPIO; aborting."
         exit 1
      }
   done
   popd >/dev/null
}

function extract_source_packages() {
   local SOURCES DEST_REPOSITORY i INSTALL_SRPMS

   [ "$1" ] && DEST_REPOSITORY=$1
   [ "$2" ] && SOURCES=$2

   [ "$DEST_REPOSITORY" ] || DEST_REPOSITORY=$MOUNTDIR/sources

   INSTALL_SRPMS=`expand_srpm_names "$SOURCES"` || exit 1

   [ -d "$DEST_REPOSITORY" ] || mkdir -p $DEST_REPOSITORY
   
   pushd $DEST_REPOSITORY >/dev/null
   for i in $INSTALL_SRPMS; do
      $RPM2CPIO $i | cpio -i --no-absolute-filenames || {
         echo "error: failed to extract $i into $DEST_REPOSITORY; aborting."
         exit 1
      }
   done
   popd >/dev/null

   for i in $INSTALL_RPMS; do
      cp $i $DEST_REPOSITORY/RPMS/$arch/ || {
         echo "error: failed to copy $i to $DEST_REPOSITORY/RPMS/$arch; aborting."
         exit 1
      }
   done

}

function add_binary_packages_to_repository_nodeps() {
   [ "$1" ] && DEST_REPOSITORY=$1
   [ "$2" ] && BINARIES=$2

   [ "$DEST_REPOSITORY" ] || DEST_REPOSITORY=$MOUNTDIR/repositories/default
   
   INSTALL_RPMS=`expand_rpm_names "$BINARIES"` || exit 1

   [ -e $DEST_REPOSITORY/RPMS/$arch ] || mkdir -p $DEST_REPOSITORY/RPMS/$arch

   for i in $INSTALL_RPMS; do
      cp $i $DEST_REPOSITORY/RPMS/$arch/ || {
         echo "error: failed to copy $i to $DEST_REPOSITORY/RPMS/$arch; aborting."
         exit 1
      }
   done

}

function add_binary_packages_to_repository() {
   local INSTALLED i INSTALL_RPMS tmpfile

   [ "$1" ] && DEST_REPOSITORY=$1
   [ "$2" ] && BINARIES=$2
   [ "$3" ] && INSTALLED="$3"

   tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX` ||
    { echo "${0##*/}: error: cannot create temporary files." >&2
       { (exit 1); exit 1; }; }

   [ "$DEST_REPOSITORY" ] || DEST_REPOSITORY=$MOUNTDIR/repositories/default

   resolve_pkg_deps "$INSTALLED" "$BINARIES"
   TOINSTALL="$PKGS_RESOLVED"

   # find again all sources also of dependency packages
   cmd="make --no-print-directory ARCH=$arch -C $GLOBALDIR SOURCESMODE=1 -s $TOINSTALL"
   [ "$VERBOSE" ] && { 
      echo 
      echo "Executing command: $cmd" >&2
   }

   BINARIES="`eval $cmd 2>$tmpfile`"

   INSTALL_RPMS=`expand_rpm_names "$TOINSTALL"` || exit 1

   [ -e $DEST_REPOSITORY/RPMS/$arch ] || mkdir -p $DEST_REPOSITORY/RPMS/$arch

   for i in $INSTALL_RPMS; do
      cp $i $DEST_REPOSITORY/RPMS/$arch/ || {
         echo "error: failed to copy $i to $DEST_REPOSITORY/RPMS/$arch; aborting."
         exit 1
      }
   done
   [ -e "$tmpfile" ] && rm -f $tmpfile
}

function add_installed_packages_to_repository() {
   local SOURCES INSTALLED i INSTALL_SRPMS INSTALL_RPMS tmpfile

   [ "$1" ] && MEDIUM_FILENAME=$1
   [ "$2" ] && DEST_REPOSITORY=$2

   [ "$MEDIUM_FILENAME" ] || {
      echo "error: add_installed_packages_to_repository: missing MEDIUM_FILENAME; aborting."
      exit 1
   }

   [ "$DEST_REPOSITORY" ] || DEST_REPOSITORY=$MOUNTDIR

   [ -e $LOCALSTATEDIR/.$MEDIUM_FILENAME.distinfo ] || {
      echo "error: file $LOCALSTATEDIR/.$MEDIUM_FILENAME.distinfo missing; aborting."
      exit 1
   }

   . $LOCALSTATEDIR/.$MEDIUM_FILENAME.distinfo

   INSTALL_SRPMS=`expand_srpm_names "$SOURCES"` || exit 1

   tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX` ||
    { echo "${0##*/}: error: cannot create temporary files." >&2
       { (exit 1); exit 1; }; }

   cmd="make --no-print-directory ARCH=$arch -C $GLOBALDIR BUILDSMODE=1 -s $SOURCES"
   [ "$VERBOSE" ] && { 
      echo 
      echo "Executing command: $cmd" >&2
   }
   BUILDS="`eval $cmd 2>$tmpfile`" || {
      echo "${0##*/}: error: the command '$cmd' exited with errors; aborting."
      exit 1
   }

   INSTALL_RPMS=`expand_rpm_names "$BUILDS"` || exit 1

   [ -e $DEST_REPOSITORY/SRPMS ] || mkdir -p $DEST_REPOSITORY/SRPMS
   [ -e $DEST_REPOSITORY/RPMS/$arch ] || mkdir -p $DEST_REPOSITORY/RPMS/$arch

   for i in $INSTALL_SRPMS; do
      cp $i $DEST_REPOSITORY/SRPMS/ || {
         echo "error: failed to copy $i to $DEST_REPOSITORY/SRPMS/; aborting."
         exit 1
      }
   done

   for i in $INSTALL_RPMS; do
      cp $i $DEST_REPOSITORY/RPMS/$arch/ || {
         echo "error: failed to copy $i to $DEST_REPOSITORY/RPMS/$arch; aborting."
         exit 1
      }
   done
   [ -e "$tmpfile" ] && rm -f $tmpfile
}

function write_installed_packages_info() {

   local SOURCES OUTLISTFILE DISTROMATICFILE tmpfile

   [ "$1" ] && MEDIUM_FILENAME=$1
   [ "$MEDIUM_FILENAME" ] || {
      echo "error: write_installed_packages_info: missing MEDIUM_FILENAME; aborting."
      exit 1
   }

   tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX` ||
    { echo "${0##*/}: error: cannot create temporary files." >&2
       { (exit 1); exit 1; }; }

   for i in $INSTALLED; do
      echo "$i" >> $tmpfile
   done

   FILESIZE=`du -h $LOCALSTATEDIR/$MEDIUM_FILENAME | awk '{print $1}'`
   echo "Target file size is $FILESIZE"

   INSTALLED=`sort $tmpfile`

   OUTLISTFILE=$LOCALSTATEDIR/.$MEDIUM_FILENAME.distinfo

   [ -e $OUTLISTFILE ] && rm -f $OUTLISTFILE
   echo "FILESIZE=\"$FILESIZE\"" > $OUTLISTFILE
   echo "INSTALLED=\"$INSTALLED\"" >> $OUTLISTFILE
   [ "$RPM_OCCUPACY_STATS_FILE" -a -e "$RPM_OCCUPACY_STATS_FILE" ] && {
      echo "RPMOCCUPACY=\"" >> $OUTLISTFILE
      cat $RPM_OCCUPACY_STATS_FILE >> $OUTLISTFILE
      echo "\"" >> $OUTLISTFILE
      rm $RPM_OCCUPACY_STATS_FILE
   }

   if [ "$INSTALLED" ]; then
      cmd="make --no-print-directory ARCH=$arch -C $GLOBALDIR SOURCESMODE=1 -s $INSTALLED"
      [ "$VERBOSE" ] && { 
         echo 
         echo "Executing command: $cmd" >&2
      }

      SOURCES="`eval $cmd >$tmpfile`"

      for i in $SOURCES; do
         echo "$i" >> $tmpfile
      done
      SOURCES=`sort $tmpfile`
   else
      SOURCES=
   fi
   echo "SOURCES=\"$SOURCES\"" >> $OUTLISTFILE

   [ "$DISTROMATIC_URL_PREFIX" ] &&
      DISTROMATICFILE=$LOCALSTATEDIR/.$MEDIUM_FILENAME.inc.html || \
      DISTROMATICFILE=""

   [ "$DISTROMATICFILE" ] && {
      cat > $DISTROMATICFILE << _EOF
<b><h1>$MAKEFILE_TARGET $PLATFORM ($arch-$LANGUAGE)</h1></b>
<p><b>Build date:</b> `date -R`
<p><b>Installed packages:</b><br>
_EOF
   }

   for i in $INSTALLED; do
        [ "$DISTROMATICFILE" ] &&
	   echo "<a href=\"${DISTROMATIC_URL_PREFIX}${i}.${arch}\">$i</a>" >> $DISTROMATICFILE

        calculate_rpms_occupation $i
#	echo "$i,$filesize" >> $OUTLISTFILE
        totalsize=`expr $totalsize + $filesize`
   done
   
   [ "$DISTROMATICFILE" ] && {
      cat >> $DISTROMATICFILE << _EOF
<br>Total RPMS size: $totalsize bytes<br>
<p><b>Source packages:</b><br>
_EOF
      for i in $SOURCES; do
        [ "$DISTROMATICFILE" ] &&
	   echo "<a href=\"${DISTROMATIC_URL_PREFIX}${i}.source\">$i</a>" >> $DISTROMATICFILE
      done
      echo "<br>" >> $DISTROMATICFILE
   }
   
   echo "INSTALLED_RPMS_SIZE=$totalsize" >> $OUTLISTFILE
   [ -e "$tmpfile" ] && rm -f $tmpfile
}

# create apt database
#
# $1: disks set list (e.g. dvd1 cd4 cd3 cd2 cd1)
# $2: current media not to add
function add_apt_medias() {
  # configure apt database for disks set (CDs or DVD)
  echo "Adding apt databases for disks set ($1).."
  CURR_MEDIA=$2
  for disknum in $1; do
     if [ -f $LOCALSTATEDIR/$DISTRONAME-$PLATFORM-$disknum.$arch.iso -a \
          "$disknum" != "$CURR_MEDIA" ]; then
        mount -o loop $LOCALSTATEDIR/$DISTRONAME-$PLATFORM-$disknum.$arch.iso $LOCALSTATEDIR/$MAKEDIST_TARGET-$PLATFORM-$disknum
        apt-cdrom add -m -d $LOCALSTATEDIR/$MAKEDIST_TARGET-$PLATFORM-$disknum \
	       -o dir::state="$MOUNTDIR/var/lib/apt" \
	       -o dir::cache="$MOUNTDIR/var/cache/apt" \
	       -o dir::etc="$MOUNTDIR/etc/apt/" > /dev/null
        umount $LOCALSTATEDIR/$DISTRONAME-$PLATFORM-$disknum.$arch.iso
     else
	echo "Warning: missing iso image $TOPDIR/$DISTRONAME-$PLATFORM-$disknum.$arch.iso"
     fi
  done
}

# compatibility function
function media_name() {
   medium_name $@
}

function medium_name() {
   local MEDIUM_NAME

   [ "$1" ] && MEDIA=$1
   [ "$2" ] && MEDIA_LABEL=$2

   MEDIUM_NAME=
   if [ "$DISTRONAME" = "${MAKEDIST_TARGET:0:${#DISTRONAME}}" ]; then
      MEDIUM_NAME=$MAKEDIST_TARGET-$PLATFORM
   else
      MEDIUM_NAME=$DISTRONAME-$MAKEDIST_TARGET-$PLATFORM
   fi
   if [ "$VARIANT" != "" ]; then
      MEDIUM_NAME=$MEDIUM_NAME-$VARIANT
   fi

   [ "$LANGUAGE" ] && MEDIUM_NAME=$MEDIUM_NAME-$LANGUAGE

   case $MEDIA in
      archive_gz|ARCHIVE_GZ)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.tar.gz
	 ;;
      archive|ARCHIVE|archive_bz2|ARCHIVE_BZ2)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.tar.bz2
	 ;;
      archive_xz|ARCHIVE_XZ)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.tar.xz
	 ;;
      cpio|CPIO)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.cpio.gz
	 ;;
      cpio_xz|CPIO_XZ)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.cpio.xz
	 ;;
      squash*|SQUASH*)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.sqsh
         ;;
      iso|ISO|iso-grub)
         [ "$MEDIA_LABEL" ] &&
	    MEDIUM_NAME=$MEDIUM_NAME-$MEDIA_LABEL.$arch.iso ||
            MEDIUM_NAME=$MEDIUM_NAME.$arch.iso
         ;;
      raw|RAW)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.img.gz
	 ;;
      qemu|QEMU)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.qcow2
         ;;
      jffs2|JFFS2)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.jffs2
         ;;
      ubifs)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.ubifs
         ;;
      ubi)
         MEDIUM_NAME=$MEDIUM_NAME.$arch.ubi
         ;;
      *)
         echo "Error: unsupported target media '$MEDIA'; aborting."
	 exit 1
	 ;;
   esac
   echo $MEDIUM_NAME
}


# produce_media()
#
# $1: media name (default: detect using medium_name function)
#
function produce_media() {

   [ "$1" ] && MEDIUM_FILENAME=$1 || MEDIUM_FILENAME=`medium_name`

   [ -e $MOUNTDIR/proc/version ] && {
      # Kill any process running in chroot
      [ "$MOUNTDIR" != "/" -a "$MOUNTDIR" != "" ] && lsof -t $MOUNTDIR | while read line; do kill -9 $line; done
      umount $MOUNTDIR/proc
      umount $MOUNTDIR/dev
      umount $MOUNTDIR/run
      umount $MOUNTDIR/sys
   }

   TARGETSIZE=`du -sh $MOUNTDIR | sed "s|[[:space:]].*||"`
   echo "Uncompressed occupacy size is: ${TARGETSIZE}"

   case $MEDIA in
      archive_gz|ARCHIVE_GZ)
         echo "Producing rootfs compressed archive $MEDIUM_FILENAME..."
         (cd $MOUNTDIR
         tar czf $LOCALSTATEDIR/$MEDIUM_FILENAME *)
	 ;;   
      archive|ARCHIVE|archive_bz2|ARCHIVE_BZ2)
         echo "Producing rootfs compressed archive $MEDIUM_FILENAME..."
         (cd $MOUNTDIR
         tar cjf $LOCALSTATEDIR/$MEDIUM_FILENAME *)
	 ;;
      archive_xz|ARCHIVE_XZ)
         echo "Producing rootfs compressed archive $MEDIUM_FILENAME..."
         (cd $MOUNTDIR
         tar cJf $LOCALSTATEDIR/$MEDIUM_FILENAME *)
	 ;;   
      cpio|CPIO)
         echo "Producing rootfs compressed cpio image $MEDIUM_FILENAME..."
         (cd $MOUNTDIR
          find . -print | cpio -o -Hnewc | gzip -c > $LOCALSTATEDIR/$MEDIUM_FILENAME )
         ;;
      cpio_xz|CPIO_XZ)
         echo "Producing rootfs compressed cpio image $MEDIUM_FILENAME..."
         (cd $MOUNTDIR
          find . -print | cpio -o -Hnewc | xz -c > $LOCALSTATEDIR/$MEDIUM_FILENAME )
         ;;
      squash|SQUASH)
         echo "Producing squashfs (default) compressed filesystem image $MEDIUM_FILENAME..."
         mksquashfs $MOUNTDIR $LOCALSTATEDIR/$MEDIUM_FILENAME -noappend
         ;;
      squash_lzma|SQUASH_LZMA)
         echo "Producing squashfs (LZMA) compressed filesystem image $MEDIUM_FILENAME..."
         mksquashfs $MOUNTDIR $LOCALSTATEDIR/$MEDIUM_FILENAME -comp lzma -noappend
         ;;
      squash_lzo|SQUASH_LZO)
         echo "Producing squashfs (LZO) compressed filesystem image $MEDIUM_FILENAME..."
         mksquashfs $MOUNTDIR $LOCALSTATEDIR/$MEDIUM_FILENAME -comp lzo -noappend
         ;;
      squash_xz|SQUASH_XZ)
         echo "Producing squashfs (XZ) compressed filesystem image $MEDIUM_FILENAME..."
         mksquashfs $MOUNTDIR $LOCALSTATEDIR/$MEDIUM_FILENAME -comp xz -noappend
         ;;
      iso|ISO)
         echo "Producing ISO image $MEDIUM_FILENAME..."
         [ -e $MOUNTDIR/boot/isolinux/isolinux.bin ] && \
            BOOTOPTIONS="-b boot/isolinux/isolinux.bin -c boot/isolinux/boot.cat -no-emul-boot -boot-load-size 4 -boot-info-table -sort $TOPDIR/mkisofs.sort" || \
            BOOTOPTIONS=""
         mkisofs -quiet \
            -o $LOCALSTATEDIR/$MEDIUM_FILENAME \
            -volid ${MAKEDIST_TARGET}_${PLATFORM} -volset $PLATFORM -publisher mambaSoft -preparer makedist \
            $BOOTOPTIONS -l -R -r $MOUNTDIR || {
            echo "Error $? returned running mkisofs; aborting."
            exit 1
         }
         isohybrid $LOCALSTATEDIR/$MEDIUM_FILENAME
         ;;
      iso-grub|ISO-GRUB)
         echo "Producing ISO-GRUB image $MEDIUM_FILENAME..."
         grub-mkrescue -o $LOCALSTATEDIR/$MEDIUM_FILENAME $MOUNTDIR -- -volid ${MAKEDIST_TARGET}_${PLATFORM} &> /dev/null
         [ $? -ne 0 ] && {
            echo "Error $? returned by grub-mkrescue; command line was:
grub-mkrescue -o $LOCALSTATEDIR/$MEDIUM_FILENAME $MOUNTDIR -- -volid ${MAKEDIST_TARGET}_${PLATFORM}"
         }
         ;;
      raw|RAW)
         echo "Producing raw image $MEDIUM_FILENAME..."
         gzip -c $LOOPDEVIMGFILE > $LOCALSTATEDIR/$MEDIUM_FILENAME
         ;;
      qemu|QEMU)
         ROOTFS_SIZE=`du -ms $MOUNTDIR | awk '{ print $1; }'`
         QEMU_SIZE=`echo "scale=0; $ROOTFS_SIZE * 1.5 / 1"| bc`
         QEMU_PART_SIZE=`echo $QEMU_SIZE - 1 | bc`
         modprobe nbd max_part=8
         echo "Producting QEMU qcow2 image $MEDIUM_FILENAME of ${QEMU_SIZE}M..."
         qemu-img create -f qcow2 $LOCALSTATEDIR/$MEDIUM_FILENAME ${QEMU_SIZE}M || {
            echo "Error: unable to create qcow2 image; aborting."
            exit 1
         }
         qemu-nbd --disconnect /dev/nbd0 || true
         qemu-nbd --connect=/dev/nbd0 $LOCALSTATEDIR/$MEDIUM_FILENAME || {
            echo "Error: unable to connect to QEMU image; aborting."
            exit 1
         }
         sleep 5
         parted -s /dev/nbd0 mklabel msdos || {
            echo "Error: unable to create partition table; aborting."
            qemu-nbd --disconnect /dev/nbd0 || true
            exit 1
         }
         parted /dev/nbd0 mkpart primary ext3 0 ${QEMU_PART_SIZE}M --align=cyl || {
            echo "Error: unable to create partition; aborting."
            qemu-nbd --disconnect /dev/nbd0 || true
            exit 1
         }
         mkfs.ext3 -q /dev/nbd0p1 || {
            echo "Error: unable to format qemu partition; aborting."
            qemu-nbd --disconnect $LOCALSTATEDIR/$MEDIUM_FILENAME
            exit 1
         }
         mount /dev/nbd0p1 $MOUNTDIR2 || {
            echo "Error: unable to mount qemu partition; aborting."
            qemu-nbd --disconnect $LOCALSTATEDIR/$MEDIUM_FILENAME
            exit 1
         }
         cp -a $MOUNTDIR/* $MOUNTDIR2 || {
            echo "Error: unable to copy rootfs data into qemu partition (disk full?); aborting."
            umount $MOUNTDIR2
            qemu-nbd --disconnect /dev/nbd0
            exit 1
         }
         umount $MOUNTDIR2
         qemu-nbd --disconnect /dev/nbd0
         ;;
      jffs2|JFFS2)
         echo "Producting jffs2 image $MEDIUM_FILENAME..."
         mkfs.jffs2 -r $MOUNTDIR -o $LOCALSTATEDIR/$MEDIUM_FILENAME -e 131072
	 ;;
      ubifs)
         echo "Producting ubifs image $MEDIUM_FILENAME..."
         mkfs.ubifs -r $MOUNTDIR -o $LOCALSTATEDIR/$MEDIUM_FILENAME -e 131072 -m 2048 -c 2047 -x favor_lzo -X 5
         # -e 131072
         ;;
      ubi)
         echo "Producting ubi image $MEDIUM_FILENAME (Flash size=${FLASH_SIZE}KiB)..."
         tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX`
         mkfs.ubifs -r $MOUNTDIR -o $tmpfile -e 131072 -m 1 -c 255 -x favor_lzo -X 5 --squash-rino-perm
         tmpfile1=`mktemp -q -t ${0##*/}.XXXXXXXX`
         cat > $tmpfile1 << _EOF
[ubifs]
mode=ubi
image=$tmpfile
vol_id=0
vol_size=${FLASH_SIZE}KiB
vol_type=dynamic
vol_name=rootfs
vol_flags=autoresize
_EOF
         ubinize -o $LOCALSTATEDIR/$MEDIUM_FILENAME -m 1 -p 128KiB $tmpfile1
         rm -f $tmpfile
         rm -f $tmpfile1
	 ;;
      *)
         echo "Error: unsupported target media '$MEDIA'; aborting."
	 exit 1
	 ;;
   esac
   write_installed_packages_info $MEDIUM_FILENAME
}

function make_rpm_first_time_script() {
   tmpfile=`mktemp -q -t ${0##*/}.XXXXXXXX`
   destfile=$MOUNTDIR/etc/rc.d/init.d/rpmscripts
   endscript=$1

   cat > $destfile << _EOF
#!/bin/bash
#
# RPM one time script executor for cross generated systems
#
echo "Performing first time system configuration..." >&2
_EOF

   $RPM -r $MOUNTDIR -qa --scripts > $tmpfile

   local section=0 ldconfig=0

   cat $tmpfile | \
   while read line; do
      case $line in
         "preuninstall scriptlet"*|"postuninstall scriptlet"*|\
         "preuninstall program"*|"postuninstall program"*) section=0; continue ;;
         "preinstall scriptlet"*|"postinstall scriptlet"*|\
         "preinstall program"*|"postinstall program"*) section=1; continue ;;
         "exit "*) continue ;;
         /sbin/ldconfig) [ $ldconfig ] || { ldconfig=1; continue; } ;;
         *) [ "$section" == "1" ] && echo "$line" >> $destfile ;;
      esac
   done
sed -i "s|exit 0||" $destfile
sed -i "s|\$1|1|" $destfile
rm -f $tmpfile
chmod +x $destfile
echo "\
$endscript
rm -f /etc/rc.d/rcsysinit.d/S99rpmscripts && {
echo 'Done. Rebooting...' >&2
sync
reboot -f
}" >> $destfile
ln -s ../init.d/rpmscripts $MOUNTDIR/etc/rc.d/rcsysinit.d/S99rpmscripts

}
